import contextlib
import html
import logging
import os

import attrs
from slack_sdk import WebClient
from slack_sdk.errors import SlackApiError

import hkkang_utils.misc as misc_utils
import hkkang_utils.socket as socket_utils

# Load environment variables
misc_utils.load_dotenv(stack_depth=2)
# Get default access token
DEFAULT_ACCESS_TOKEN = (
    os.environ["SLACK_ACCESS_TOKEN"] if "SLACK_ACCESS_TOKEN" in os.environ else None
)

logger = logging.getLogger("SlackMessenger")


@attrs.define
class SlackMessenger:
    """Note that the default token is set by the environment variable SLACK_ACCESS_TOKEN.

    Example:
        messenger = SlackMessenger(channel="test-channel")
        messenger.send_message("Hello World")
    """

    channel: str = attrs.field()
    token: str = attrs.field(default=attrs.Factory(lambda: DEFAULT_ACCESS_TOKEN))
    append_src_info: bool = attrs.field(default=True)

    def __attrs_post_init__(self):
        if self.token is None:
            raise ValueError(
                """Please set token or set SLACK_ACCESS_TOKEN environment variable.
                    If you don't have the access token, follow the tutorial 
                    to get bot OAuthToken and setup the bot permissions. 
                    https://github.com/slackapi/python-slack-sdk/tree/main/tutorial"""
            )

    def send(self, text: str) -> None:
        """Send message to slack channel

        :param text: Message to send
        :type text: str
        """
        return send_message(
            token=self.token,
            channel=self.channel,
            text=text,
            append_src_info=self.append_src_info,
        )


def send_message(
    channel: str,
    text: str,
    token: str = DEFAULT_ACCESS_TOKEN,
    append_src_info: bool = True,
) -> None:
    """Send message to Slack channel

    Example:
        send_message(channel="test-channel", text="Hello world")

    :param channel: Name of the channel to send the message
    :type channel: str
    :param text: Message to send
    :type text: str
    :param token: Slack access token, defaults to DEFAULT_ACCESS_TOKEN
    :type token: str, optional
    :param append_src_info: Whether to tell which part of the code the function is called, defaults to True
    :type append_src_info: bool, optional
    """
    # Check if token is provided
    if token is None:
        raise ValueError(
            "Please set token or set SLACK_ACCESS_TOKEN environment variable."
        )
    # Create client
    client = WebClient(token=token)

    # Build message
    if append_src_info:
        ip = socket_utils.get_local_ip()
        host_name = socket_utils.get_host_name()
        text_with_prefix = f"Message from {host_name}({ip}):\n{text}"

    # Send message
    try:
        response = client.chat_postMessage(channel=channel, text=text_with_prefix)
        decoded_text = html.unescape(response["message"]["text"])
        assert decoded_text == text_with_prefix, f"{decoded_text}"
        logger.info(f"Sending message to channel {channel}: {text}")
    except SlackApiError as e:
        # You will get a SlackApiError if "ok" is False
        assert e.response["ok"] is False
        assert e.response["error"], "channel_not_found"
        logger.info(f"Got an error: {e.response['error']}")


@contextlib.contextmanager
def notification(
    channel: str,
    success_msg: str = None,
    error_msg: str = None,
    token: str = DEFAULT_ACCESS_TOKEN,
    disable: bool = False,
) -> None:
    """Send message when the task within the code block is finished

    Example:
        import hkkang_utils.slack as slack_utils

        with slack_utils.notification(
            channel="test-channel",
            success_msg="Process done!",
            error_msg="Error raised during the process"
        ):
            # Perform your task here
            ...

    :param channel: Name of the channel to send the message
    :type channel: str
    :param success_msg: Message to send when the given code block completes, defaults to None
    :type success_msg: str, optional
    :param error_msg: Message to send when error raise with the given code block, defaults to None
    :type error_msg: str, optional
    :param token: slack access token, defaults to DEFAULT_ACCESS_TOKEN
    :type token: str, optional
    :rtype: None
    """
    if misc_utils.is_debugger_active():
        disable = True

    if disable:
        yield None
        return None

    slack_messenger = SlackMessenger(channel=channel, token=token)
    try:
        yield slack_messenger
        if success_msg is not None:
            slack_messenger.send(success_msg)
    except Exception as e:
        if error_msg is None:
            message_to_send = f"Error occurred at {e.__class__.__name__}: {e}"
        else:
            message_to_send = f"{error_msg} ({e.__class__.__name__}: {e})"
        slack_messenger.send(message_to_send)
        raise e


@contextlib.contextmanager
def slack_notification(*args, **kwargs):
    raise NotImplementedError("Please use notification instead of slack_notification")



import unittest

import src.hkkang_utils.misc as misc_utils


class Test_misc_utils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_misc_utils, self).__init__(*args, **kwargs)
    
    def test_infinit_iterator(self):
        input_list = [1,2,3,4,5]
        result = []
        for i, item in enumerate(misc_utils.infinite_iterator(input_list)):
            if i == 10:
                break
            result.append(item)
        gold = [1,2,3,4,5,1,2,3,4,5]
        self.assertEqual(result, gold, f"result: {result}, gold: {gold}")
    
    def test_to_dict(self):
        class TmpObject():
            def __init__(self, a,b,c,d,e):
                self.a=a
                self.b=b
                self._c=c
                self.__d=d
                self.private_e=e
        tmp = TmpObject(1,2,3,4,5)   
        tmp_dict = misc_utils.to_dict(tmp, exclude_prefixes=["_", "__", "private"]) 
        self.assertEqual(type(tmp_dict), dict)
        self.assertDictEqual(tmp_dict, {"a":1, "b":2})
    
if __name__ == "__main__":
    unittest.main()
    
    


import re
import nltk
import random
import string

try:
    STOPWORDS = set(nltk.corpus.stopwords.words('english'))
except:
    nltk.download("stopwords")
    nltk.download('punkt')
    STOPWORDS = set(nltk.corpus.stopwords.words('english'))

BOOLEAN_STRINGS = ["true", "false", "yes", "no", "1", "0"]

def multi_space_to_single_space(text: str) -> str:
    return re.sub(' +', ' ', text)

def is_int(text: str) -> bool:
    try:
        int(text)
        return True
    except ValueError:
        return False
    
def is_number(string: str) -> bool:
    try:
        float(string)
        return True
    except ValueError:
        return False
    
def generate_random_str(size: int=6, chars=string.ascii_uppercase + string.digits)-> str:
    return ''.join(random.choice(chars) for _ in range(size))

def remove_substring(text: str, sub_str: str) -> str:
    # The thrid argument of str.maketrans is a string of characters to be removed
    return text.replace(sub_str, "")

def remove_characters(text: str, characters: str) -> str:
    return text.translate(str.maketrans('', '', characters))

def remove_punctuation(text: str) -> str:
    return remove_characters(text, string.punctuation)


import unittest

import src.hkkang_utils.tensor as tensor_utils
from tests.test_tensor.utils import (
    Float16ModelWithBias,
    Float32ModelWithBias,
    Float32ModelWithOutBias,
    to_mb,
)


class Test_tensor_utils(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(Test_tensor_utils, self).__init__(*args, **kwargs)

    def test_zero_pad_batching_one_dim(self):
        pass

    def test_zero_pad_batching_two_dim(self):
        pass

    def test_zero_pad_batching(self):
        pass

    def test_calculate_model_size(self):
        model_f32_wo_bias = Float32ModelWithOutBias()
        model_size = tensor_utils.calculate_model_size(model_f32_wo_bias)
        self.assertEqual(model_size, to_mb(model_f32_wo_bias.model_size))

        model_f32_w_bias = Float32ModelWithBias()
        model_size = tensor_utils.calculate_model_size(model_f32_w_bias)
        self.assertEqual(model_size, to_mb(model_f32_w_bias.model_size))

        model_f16_w_bias = Float16ModelWithBias()
        model_size = tensor_utils.calculate_model_size(model_f16_w_bias)
        self.assertEqual(model_size, to_mb(model_f16_w_bias.model_size))


if __name__ == "__main__":
    unittest.main()



import torch

FEAT_SIZE = 10
FLOAT32_SIZE = 4
FLOAT16_SIZE = 2


def to_mb(size: int) -> float:
    return size / 1024 / 1024


class Float32ModelWithBias(torch.nn.Module):
    def __init__(self):
        super(Float32ModelWithBias, self).__init__()
        self.linear1 = torch.nn.Linear(
            FEAT_SIZE, FEAT_SIZE, bias=True, dtype=torch.float32
        )
        self.linear2 = torch.nn.Linear(
            FEAT_SIZE, FEAT_SIZE, bias=True, dtype=torch.float32
        )
        self.linear3 = torch.nn.Linear(
            FEAT_SIZE, FEAT_SIZE, bias=True, dtype=torch.float32
        )

    def forward(self, x):
        return self.linear3(self.linear2(self.linear1(x)))

    @property
    def layer_num(self):
        return 3

    @property
    def weight_size(self):
        return (FEAT_SIZE * FEAT_SIZE) * self.layer_num

    @property
    def bias_size(self):
        return FEAT_SIZE * self.layer_num

    @property
    def model_size(self):
        return (self.weight_size + self.bias_size) * FLOAT32_SIZE


class Float32ModelWithOutBias(torch.nn.Module):
    def __init__(self):
        super(Float32ModelWithOutBias, self).__init__()
        self.linear1 = torch.nn.Linear(
            FEAT_SIZE, FEAT_SIZE, bias=False, dtype=torch.float32
        )
        self.linear2 = torch.nn.Linear(
            FEAT_SIZE, FEAT_SIZE, bias=False, dtype=torch.float32
        )
        self.linear3 = torch.nn.Linear(
            FEAT_SIZE, FEAT_SIZE, bias=False, dtype=torch.float32
        )

    def forward(self, x):
        return self.linear3(self.linear2(self.linear1(x)))

    @property
    def layer_num(self):
        return 3

    @property
    def weight_size(self):
        return (FEAT_SIZE * FEAT_SIZE) * self.layer_num

    @property
    def bias_size(self):
        return 0 * self.layer_num

    @property
    def model_size(self):
        return (self.weight_size + self.bias_size) * FLOAT32_SIZE


class Float16ModelWithBias(torch.nn.Module):
    def __init__(self):
        super(Float16ModelWithBias, self).__init__()
        self.linear1 = torch.nn.Linear(
            FEAT_SIZE, FEAT_SIZE, bias=True, dtype=torch.float16
        )
        self.linear2 = torch.nn.Linear(
            FEAT_SIZE, FEAT_SIZE, bias=True, dtype=torch.float16
        )
        self.linear3 = torch.nn.Linear(
            FEAT_SIZE, FEAT_SIZE, bias=True, dtype=torch.float16
        )

    def forward(self, x):
        return self.linear3(self.linear2(self.linear1(x)))

    @property
    def layer_num(self):
        return 3

    @property
    def weight_size(self):
        return (FEAT_SIZE * FEAT_SIZE) * self.layer_num

    @property
    def bias_size(self):
        return FEAT_SIZE * self.layer_num

    @property
    def model_size(self):
        return (self.weight_size + self.bias_size) * FLOAT16_SIZE



import abc
from typing import Callable, Dict


# This file containes design patterns
class SingletonMeta(type):
    """Meta Singleton class"""

    _instances: Dict[str, Callable] = dict()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances.keys():
            cls._instances[cls] = super(SingletonMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class SingletonABCMeta(abc.ABCMeta):
    """Abstract and meta Singleton class"""

    _instances: Dict[str, Callable] = dict()

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances.keys():
            cls._instances[cls] = super(SingletonABCMeta, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


# Decorator
def singleton(cls):
    """Singleton decorator"""
    instance: Dict[str, Callable] = None

    def wrapper(*args, **kwargs):
        nonlocal instance
        if instance is None:
            instance = cls(*args, **kwargs)
        return instance

    return wrapper


class SingletonMetaWithArgs(type):
    """Meta Singleton class with arguments. One object is created for each set of arguments"""

    _instances: Dict[str, Dict[str, Callable]] = dict()

    def __call__(cls, *args, **kwargs):
        # Get instance key
        instance_key = cls.__repr_args__(*args, **kwargs)
        if cls not in cls._instances.keys():
            cls._instances[cls] = dict()
        # Create instance if not exists
        if instance_key not in cls._instances[cls].keys():
            cls._instances[cls][instance_key] = super(
                SingletonMetaWithArgs, cls
            ).__call__(*args, **kwargs)
        return cls._instances[cls][instance_key]

    @staticmethod
    def __repr_args__(*args, **kwargs):
        return str(args) + str(kwargs)


class SingletonABCMetaWithArgs(abc.ABCMeta):
    """Abstract and Meta Singleton class with arguments. One object is created for each set of arguments"""

    _instances: Dict[str, Dict[str, Callable]] = dict()

    def __call__(cls, *args, **kwargs):
        # Get instance key
        instance_key = cls.__repr_args__(*args, **kwargs)
        if cls not in cls._instances.keys():
            cls._instances[cls] = dict()
        # Create instance if not exists
        if instance_key not in cls._instances[cls].keys():
            cls._instances[cls][instance_key] = super(
                SingletonABCMetaWithArgs, cls
            ).__call__(*args, **kwargs)
        return cls._instances[cls][instance_key]

    @staticmethod
    def __repr_args__(*args, **kwargs):
        return str(args) + str(kwargs)


# Decorator
def singletonWithArgs(cls):
    """SingletonWithArgs decorator"""
    instances: Dict = dict()

    def wrapper(*args, **kwargs):
        nonlocal instances
        instance_key = SingletonMetaWithArgs.__repr_args__(*args, **kwargs)
        if instance_key not in instances.keys():
            instances[instance_key] = cls(*args, **kwargs)
        return instances[instance_key]

    return wrapper
